import datetime
from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django.utils.html import conditional_escape, mark_safe
from django.utils.timezone import make_aware
from django.contrib.auth import get_user_model
from . import models

class FristAuswahlForm(forms.Form):
    def __init__(self, fristen, anfang, *args, **kwargs):
        super().__init__(*args, **kwargs)
        choices=[(frist.pk, frist.kurz) for frist in fristen]
        self.fields['frist_auswahl'] = forms.ChoiceField(
            choices=choices,
            initial=anfang
        )

def datum_iterator(zeit_anfang, zeit_ende):
    delta = zeit_ende - zeit_anfang
    for day in range(delta.days+1):
        yield zeit_anfang + datetime.timedelta(day)

class KalenderForm(forms.Form):
    frist = None
    zuweisung = None
    ist_abgeschlossen = True

    def __init__(self, frist, zuweisung, ist_abgeschlossen, *args, **kwargs):
        kwargs.setdefault('label_suffix', '')
        super().__init__(*args, **kwargs)

        if frist is None:
            raise ValidationError(
                _('Keine Frist angegeben'),
                code='keine Frist'
            )
        self.frist = frist

        if zuweisung is None:
            raise ValidationError(
                    _('Keine Projektzuweisug gegeben'),
                    code='keine Zuweisung'
                )
        self.zuweisung = zuweisung

        self.ist_abgeschlossen = ist_abgeschlossen

        anfang = self.frist.zeitraum_anfang
        ende = self.frist.zeitraum_ende
        heute = datetime.date.today()
        for datum in datum_iterator(anfang, ende):
            if datum > heute or self.ist_abgeschlossen:
                disabled = True
            else:
                disabled = False
            self.fields[str(datum)] = forms.DecimalField(
                required=False,
                label=str(datum.day),
                disabled=disabled,
                max_digits=4,
                widget=forms.NumberInput(attrs={
                    'min' : 0,
                    'max' : 24,
                    'step' : 0.01
                })
            )

    def als_html_tabelle(self):
        output = [
            '<table>',
            '  <thead>',
            '      <tr><th colspan="1"></th>',
            '      <th>Mo</th>',
            '      <th>Di</th>',
            '      <th>Mi</th>',
            '      <th>Do</th>',
            '      <th>Fr</th>',
            '      <th>Sa</th>',
            '      <th>So</th>',
            '   </tr>',
            '  </thead>',
            '  <tbody>',]

        # Formularfehler
        top_errors = self.non_field_errors().copy()
        if top_errors:
            output.insert(0, '%s' % top_errors)

        # Muster fuer Formularzellen
        zelle = '%(begin_tag)s%(label)s%(field)s%(help_text)s%(errors)s%(end_tag)s'

        erste = True
        anfang = self.frist.zeitraum_anfang
        ende = self.frist.zeitraum_ende

        for datum in datum_iterator(anfang, ende):
            bound_field = self[str(datum)]

            # Erstelle errors
            bf_errors = self.error_class(bound_field.errors)

            # Erstelle label
            if bound_field.label:
                label = conditional_escape(bound_field.label)
                label = bound_field.label_tag(label) or ''
            else:
                label = ''

            # Erstelle help_text
            if self.fields[str(datum)].help_text:
                help_text = '<br><span class="helpt' \
                    'ext">%s</span>' % self.fields[str(datum)].help_text
            else:
                help_text = ''

            # Erstelle begin_tag
            iso = datum.isocalendar()
            if erste and iso[2] == 1:
                begin_tag = '    <tr class="kw"><th>KW %s</th>\n<td>' % iso[1]
                erste = False
            elif erste:
                begin_tag = '    <tr class="kw"><th>KW' \
                    '%s</th>\n<td colspan="%s"></td><td>' % (iso[1], iso[2]-1)
                erste = False
            elif iso[2] == 1:
                begin_tag = '    <tr class="kw"><th>KW %s</th><td>' % iso[1]
            else:
                begin_tag = '    <td>'

            # Erstelle end_tag
            if iso[2] == 7:
                end_tag = '</td>\n</tr>'
            elif (datum + datetime.timedelta(1)).month > datum.month:
                end_tag = '</td><td colspan="%s"></td>\n</tr>' % (7 - iso[2])
                erste = True
            else:
                end_tag = '</td>'

            # Fuege alles Zusammen zur Formularzelle
            output.append(zelle % {
                    'begin_tag' : begin_tag,
                    'errors': bf_errors,
                    'label': label,
                    'field': bound_field,
                    'help_text': help_text,
                    'field_name': bound_field.html_name,
                    'end_tag' : end_tag,
                }
            )

        # Ende der Tabelle
        output.append('  </tbody>\n</table>')

        return mark_safe('\n'.join(output))

    def clean(self):
        super().clean()

        geaendert = len(self.changed_data) != 0
        if self.ist_abgeschlossen and geaendert:
            raise ValidationError(
                _('Die Frist wurde bereits abgeschlossen'),
                code='abgeschlossen'
            )

        jetzt = make_aware(datetime.datetime.now())
        if self.frist.frist_ende < jetzt and geaendert:
            raise ValidationError(
                _('Die Frist zur Eintragung wurde überschritten.'),
                code='fristende'
            )

        anfang = self.frist.zeitraum_anfang
        ende = self.frist.zeitraum_ende
        heute = datetime.date.today()
        for datum in datum_iterator(anfang, ende):
            feld = str(datum)
            arbeitszeit = self.cleaned_data.get(feld)
            if arbeitszeit is not None and arbeitszeit < 0:
                fehler = ValidationError(
                    _('Zu wenig Arbeitsstunden eingetragen'),
                    code='mininmale Arbeitsstunden'
                )
                self.add_error(feld, fehler)
            if arbeitszeit is not None and arbeitszeit > 24:
                fehler = ValidationError(
                    _('Zu viele Arbeitsstunden eingetragen'),
                    code='maximale Arbeitsstunden'
                )
                self.add_error(feld, fehler)
            ist_zukunft = datum > heute
            if arbeitszeit is not None and ist_zukunft:
                fehler = ValidationError(
                    _('Das Datum der Eintragun liegt in der Zukunft'),
                    code='zukunft'
                )
                self.add_error(feld, fehler)
        return self.cleaned_data

    def save(self):
        eintraege = models.Stundenkonto.objects.filter(
            projekt_zuweisung=self.zuweisung,
            frist=self.frist,
        )
        for feld_name in self.changed_data:
            datum = datetime.datetime.strptime(
                feld_name,
                '%Y-%m-%d'
            ).date()
            arbeitszeit = self.cleaned_data.get(feld_name)
            try:
                eintrag = eintraege.get(
                    arbeits_tag=datum
                )
                if arbeitszeit is None or arbeitszeit == 0:
                    eintrag.delete()
                else:
                    eintrag.arbeits_zeit = arbeitszeit
                    eintrag.save()
            except models.Stundenkonto.DoesNotExist:
                if arbeitszeit is not None and arbeitszeit != 0:
                    eintrag = models.Stundenkonto(
                        projekt_zuweisung=self.zuweisung,
                        frist=self.frist,
                        arbeits_tag=datum,
                        arbeits_zeit=arbeitszeit,
                    )
                    eintrag.save()

class AbschliessenForm(forms.Form):
    abschliessen = forms.BooleanField(
        required=True,
        widget=forms.HiddenInput,
        initial=True
    )

class SuchForm(forms.Form):
    suchfeld = forms.CharField(
        max_length=511,
        required=True,
        widget=forms.TextInput(attrs={'autocomplete': 'off'})
    )

    def clean_suchfeld(self):
        eingabe = iter(self.cleaned_data.get('suchfeld'))
        filter_func = lambda c: c.isalnum() or c == ' ' or c == '@'
        return ''.join(filter(filter_func, eingabe))

class MitarbeiterInSuchForm(SuchForm):
    pass

class MitarbeiterRegistrierungForm(forms.ModelForm):
    class Meta:
        model = get_user_model()
        fields = ['vorname', 'nachname', 'email']

    def clean_email(self):
        email = self.cleaned_data.get('email')
        if not email:
            raise ValueError(
                _('Emailadresse nicht angegeben'),
                code='email fehlt'
            )
        email = self.Meta.model.objects.normalize_email(email)
        return email

    def save(self, commit=True):
        neue_nutzer_in = super().save(commit=False)
        neue_nutzer_in.set_unusable_password()
        if commit:
            neue_nutzer_in.save()
        return neue_nutzer_in

class ProjektMitarbeiterRegistrierungForm(MitarbeiterRegistrierungForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        stufen = models.Gehaltsstufe.objects.all()
        choices = ((stufe.pk, stufe.stufen_kuerzel) for stufe in stufen)
        self.fields['stufen_auswahl'] = forms.ChoiceField(
            required=True,
            choices=choices
        )

class MitarbeiterInGehaltsstufenForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        stufen = models.Gehaltsstufe.objects.all()
        choices = ((stufe.pk, stufe.stufen_kuerzel) for stufe in stufen)
        self.fields['stufen_auswahl'] = forms.ChoiceField(
            required=True,
            choices=choices
        )

class DateInput(forms.DateInput):
    input_type = 'date'

    def format_value(self, value):
        return value

class DateTimeInput(forms.DateTimeInput):
    input_type = 'datetime-local'

    def format_value(self, value):
        return value

class ProjektAendernForm(forms.ModelForm):
    class Meta:
        model = models.Projekt
        fields = [
            'projekt_name',
            'auftraggeber',
            'laufzeit_anfang',
            'laufzeit_ende',
        ]

    def clean(self, *args, **kwargs):
        cleaned_data = super().clean(*args, **kwargs)
        anfang = cleaned_data.get('laufzeit_anfang')
        ende = cleaned_data.get('laufzeit_ende')
        if anfang > ende:
            raise ValidationError(
                _('Laufzeit ende kann nicht vor Laufzeit anfang liegen!'),
                code='laufzeit'
            )
        return cleaned_data

class ProjektErstellenForm(ProjektAendernForm):
    class Meta:
        model = models.Projekt
        fields = [
            'projekt_name',
            'auftraggeber',
            'kostenstelle',
            'laufzeit_anfang',
            'laufzeit_ende',
        ]

class ProjektEinrichtungsAuswahlForm(forms.Form):
    def __init__(self, query_set, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['einrichtungen'] = forms.ChoiceField(
            choices=((
                    einrichtung.pk, 
                    einrichtung.einrichtungs_name
                ) for einrichtung in query_set),
            required=True,
            widget=forms.RadioSelect
        )

class ProjektZuweisungForm(forms.ModelForm):

    class Meta:
        model = models.ProjektZuweisung
        fields = ['hat_gehalts_stufe']
        labels = {
            'hat_gehalts_stufe': 'Gehaltsstufe',
        }

class ProjektMitarbeiterForm(forms.ModelForm):
    class Meta:
        model = get_user_model()
        fields = ['vorname', 'nachname']

class Stundenzettel(forms.Form):
    projekt = None
    frist = None
    zuweisungen = None
    abgeschlossene = None
    zeitraum = None
    stunden = None

    def __init__(self, projekt, frist, ist_verwalter=False, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.projekt = projekt
        self.frist = frist
        self.ist_verwalter = ist_verwalter

        self.zuweisungen = models.ProjektZuweisung.objects.filter(
            projekt=self.projekt
        )

        self.abgeschlossene = models.Abgeschlossen.objects.filter(
            projekt_zuweisung__projekt=projekt,
            frist=self.frist
        ).select_related('projekt_zuweisung')
        self.stunden = models.Stundenkonto.objects.filter(
            projekt_zuweisung__projekt=projekt,
            frist=self.frist
        ).select_related('projekt_zuweisung')

        anfang = self.frist.zeitraum_anfang
        ende = self.frist.zeitraum_ende
        self.zeitraum = list(datum_iterator(anfang, ende))
        heute = datetime.date.today()
        jetzt = make_aware(datetime.datetime.now())

        for zuweisung in self.zuweisungen:
            for datum in self.zeitraum:
                if self.ist_verwalter:
                    außerhalb_erlaubt = datum > heute
                else:
                    außerhalb_erlaubt = datum > heute or self.frist.frist_ende < jetzt
                self.fields[str(zuweisung.id) + '-' + str(datum)] = forms.DecimalField(
                    required=False,
                    label=str(datum.day),
                    max_digits=4,
                    disabled=außerhalb_erlaubt,
                    widget=forms.NumberInput(attrs={
                        'min' : 0,
                        'max' : 24,
                        'step' : 0.01
                    })
                )

        for abgeschlossen in self.abgeschlossene:
            ist_abgeschlossen = abgeschlossen.ist_abgeschlossen
            for datum in self.zeitraum:
                feld = self.fields.get(str(abgeschlossen.projekt_zuweisung.id) + '-' +
                        str(datum))
                if feld is not None:
                    feld.disabled = feld.disabled or ist_abgeschlossen

        for eintrag in self.stunden:
            feld = self.fields.get(str(eintrag.projekt_zuweisung.id) + '-' +
                    str(eintrag.arbeits_tag))
            if feld is not None:
                feld.initial = eintrag.arbeits_zeit

    def als_html_tabelle(self):
        # Kopf der Tabelle
        output = [
            '<table>',
            '    <thead>',
            '        <tr>',
            '            <th>Nachname, Vorname</th>',
        ]

        for datum in self.zeitraum:
            output.append('            <th>%s</th>' % (datum.day))

        output += [
            '        </tr>',
            '    </thead>',
            '    <tbody>',
        ]

        # Formularfehler
        top_errors = self.non_field_errors().copy()
        if top_errors:
            output.insert(0, '%s' % top_errors)

        # Muster fuer Formularzellen
        zelle = '            <td>%(label)s%(field)s%(help_text)s%(errors)s</td>'

        # Reihen
        for zuweisung in self.zuweisungen:
            reihe = [
                '    <tr>',
                '        <td class="name">%s, %s</td>' % (
                zuweisung.nutzer_in.nachname,
                zuweisung.nutzer_in.vorname)
            ]

            for datum in self.zeitraum:
                bound_field = self[str(zuweisung.id) + '-' + str(datum)]
                feld = self.fields[str(zuweisung.id) + '-' + str(datum)]

                # Erstelle errors
                bf_errors = self.error_class(bound_field.errors)

                # Erstelle label
                if bound_field.label:
                    label = conditional_escape(bound_field.label)
                    label = bound_field.label_tag(label) or ''
                else:
                    label = ''

                # Erstelle help_text
                if feld.help_text:
                    help_text = '<br><span class="helpt' \
                        'ext">%s</span>' % feld.help_text
                else:
                    help_text = ''

                # Fuege alles Zusammen zur Formularzelle
                reihe.append(zelle % {
                        'errors': bf_errors,
                        'label': label,
                        'field': bound_field,
                        'help_text': help_text,
                        'field_name': bound_field.html_name,
                    }
                )

            reihe.append('        </tr>')
            output += reihe

        output += [
            '    </tbody>'
            '</table>'
        ]

        return mark_safe('\n'.join(output))


    def clean(self):
        cleaned_data = super().clean()
        heute = datetime.date.today()
        jetzt = make_aware(datetime.datetime.now())
        
        if jetzt > self.frist.frist_ende and not self.ist_verwalter:
            raise ValidationError(
                _('Der Eintragunszeitraum ist bereist verstrichen'),
                code='zeitraum'
            )

        for zuweisung in self.zuweisungen:

            abgeschlossen = self.abgeschlossene.filter(
                        projekt_zuweisung=zuweisung
                    ).first()
            ist_abgeschlossen = abgeschlossen is not None
            ist_abgeschlossen = ist_abgeschlossen and abgeschlossen.ist_abgeschlossen

            for datum in self.zeitraum:
                feld_name = str(zuweisung.id) + '-' + str(datum)
                arbeitszeit = cleaned_data.get(feld_name) or 0
                geaendert = feld_name in self.changed_data
                if geaendert and arbeitszeit < 0:
                    self.add_error(
                        feld_name,
                        ValidationError(
                            _('Zu wenig Arbeitsstunden eingetragen'),
                            code='mininmale Arbeitsstunden'
                        )
                    )
                elif geaendert and arbeitszeit > 24:
                    self.add_error(
                        feld_name,
                        ValidationError(
                            _('Zu viele Arbeitsstunden eingetragen'),
                            code='mininmale Arbeitsstunden'
                        )
                    )
                elif geaendert and heute < datum:
                    self.add_error(
                        feld_name,
                        ValidationError(
                        _('Das Datum der Eintragun liegt in der Zukunft'),
                        code='zukunft'
                        )
                    )
                elif geaendert and ist_abgeschlossen:
                    self.add_error(
                        feld_name,
                        ValidationError(
                        _('Die Frist ist bereits abgeschlossen'),
                        code='abgeschlossen'
                        )
                    )
        return cleaned_data

    def save(self):
        for zuweisung in self.zuweisungen:
            stunden = self.stunden.filter(projekt_zuweisung=zuweisung)
            for datum in self.zeitraum:
                feld_name = str(zuweisung.id) + '-' + str(datum)
                geaendert = feld_name in self.changed_data
                arbeitszeit = self.cleaned_data.get(feld_name)
                if geaendert:
                    try:
                        eintrag = stunden.get(
                            arbeits_tag=datum
                        )
                        if arbeitszeit is None or arbeitszeit == 0:
                            eintrag.delete()
                        else:
                            eintrag.arbeits_zeit = arbeitszeit
                            eintrag.save()
                    except models.Stundenkonto.DoesNotExist:
                        if arbeitszeit is not None and arbeitszeit != 0:
                            eintrag = models.Stundenkonto(
                                projekt_zuweisung=zuweisung,
                                frist=self.frist,
                                arbeits_tag=datum,
                                arbeits_zeit=arbeitszeit,
                            )
                            eintrag.save()

class ProjektLeiterInForm(forms.ModelForm):
    class Meta:
        model = models.ProjektLeiterIn
        fields = [
            'nutzer_in',
            'leitet',
            'von',
            'bis',
            'ist_vertreter',
        ]
        widgets = {
            'nutzer_in' : forms.HiddenInput(),
            'leitet' : forms.HiddenInput(),
            'von' : DateInput(),
            'bis' : DateInput()
        }

    def clean(self, *args, **kwargs):
        cleaned_data = super().clean(*args, **kwargs)
        anfang = cleaned_data.get('von')
        ende = cleaned_data.get('bis')
        if anfang > ende:
            raise ValidationError(
                _('Laufzeit ende kann nicht vor Laufzeit anfang liegen!'),
                code='laufzeit'
            )
        return cleaned_data

class VerwaltungProjektMitarbeiterInForm(forms.ModelForm):
    class Meta:
        model = models.ProjektZuweisung
        fields = ['nutzer_in', 'projekt', 'hat_gehalts_stufe']
        widgets = {
            'nutzer_in' : forms.HiddenInput(),
            'projekt' : forms.HiddenInput()
        }
