import datetime
from django.db import models
from django.conf import settings
from django.utils.timezone import make_aware, template_localtime
from django.utils.formats import date_format

# Create your models here.

class Einrichtung(models.Model):
    einrichtungs_name = models.CharField(
        max_length=255,
        unique=True
    )
    verbuchungs_stelle = models.CharField(
        max_length=7,
        unique=True
    )

    def __str__(self):
        return self.einrichtungs_name

class Projekt(models.Model):
    kostenstelle = models.IntegerField(unique=True)
    projekt_name = models.CharField(
        max_length=255
    )
    auftraggeber = models.CharField(
        max_length=255
    )
    laufzeit_anfang = models.DateField()
    laufzeit_ende = models.DateField()
    einrichtung = models.ForeignKey(
        'Einrichtung',
        on_delete=models.CASCADE,
        null=True
    )

    class Meta:
        constraints = [
            models.CheckConstraint(
                check=models.Q(laufzeit_ende__gt=models.F('laufzeit_anfang')),
                name='laufzeit_check_projekt'
            )
        ]

    def __str__(self):
        return self.projekt_name

class ProjektLeiterIn(models.Model):
    nutzer_in = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.RESTRICT
    )
    leitet = models.ForeignKey(
        'Projekt',
        on_delete=models.CASCADE
    )
    von = models.DateField()
    bis = models.DateField()
    ist_vertreter = models.BooleanField(
    )

    class Meta:
        constraints = [
            models.CheckConstraint(
                check=models.Q(bis__gt=models.F('von')),
                name='laufzeit_check_leitung'
            )
        ]
        unique_together = ['nutzer_in', 'leitet']

class Gehaltsstufe(models.Model):
    stufen_kuerzel = models.CharField(
        max_length=255,
        unique=True
    )

    def __str__(self):
        return self.stufen_kuerzel

class ProjektZuweisung(models.Model):
    nutzer_in = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.RESTRICT,
        related_name='projektzuweisung'
    )
    hat_gehalts_stufe = models.ForeignKey(
        'Gehaltsstufe',
        on_delete=models.RESTRICT
    )
    projekt = models.ForeignKey(
        'Projekt',
        on_delete=models.RESTRICT
    )

    class Meta:
        unique_together = ['projekt', 'nutzer_in']

class FristManager(models.Manager):
    def get_queryset(self):
        self._update()
        return super().get_queryset()

    def _update(self):
        heute = datetime.date.today()
        laufend = super().get_queryset().filter(zeitraum_ende__gte=heute)
        if not laufend.exists():
            letzte = super().get_queryset().order_by('-zeitraum_ende').first()
            try:
                anfang = letzte.zeitraum_ende + datetime.timedelta(1)
            except AttributeError:
                anfang = datetime.date(year=2021, month=1, day=1)

            if anfang.month == 12:
                jahr_z_ende = anfang.year + 1
                jahr_f_ende = anfang.year + 1
            elif anfang.month == 11:
                jahr_z_ende = anfang.year
                jahr_f_ende = anfang.year + 1
            else:
                jahr_z_ende = anfang.year
                jahr_f_ende = anfang.year

            z_ende = datetime.date(
                year = jahr_z_ende,
                month = (anfang.month % 12) + 1,
                day = 1
            )
            z_ende = z_ende - datetime.timedelta(1)
            f_ende = datetime.datetime(
                year = jahr_f_ende,
                month = ((anfang.month + 1) % 12) + 1,
                day = 1
            )
            f_ende = make_aware(f_ende - datetime.timedelta(microseconds=1))
            super().get_queryset().create(
                frist_ende = f_ende,
                zeitraum_anfang = anfang,
                zeitraum_ende = z_ende
            )
            return self._update()
        else:
            return True

class Frist(models.Model):
    frist_ende = models.DateTimeField()
    zeitraum_anfang = models.DateField()
    zeitraum_ende = models.DateField()
    objects = FristManager()

    class Meta:
        ordering = ['-zeitraum_ende']

    def __str__(self):
        return '%s - %s : %s' % (self.zeitraum_anfang, self.zeitraum_ende, self.frist_ende)

    def kurz(self):
        anfang = date_format(self.zeitraum_anfang, format='SHORT_DATE_FORMAT', use_l10n=True)
        ende = date_format(self.zeitraum_ende, format='SHORT_DATE_FORMAT', use_l10n=True)
        bis = date_format(template_localtime(self.frist_ende), format='SHORT_DATETIME_FORMAT', use_l10n=True)
        return '%s - %s ( %s )' % (anfang, ende, bis)

class Stundenkonto(models.Model):
    projekt_zuweisung = models.ForeignKey(
        'ProjektZuweisung',
        on_delete=models.RESTRICT
    )
    frist = models.ForeignKey(
        'Frist',
        on_delete=models.RESTRICT
    )
    arbeits_tag = models.DateField()
    arbeits_zeit = models.DecimalField(
        decimal_places=2,
        max_digits=4
    )

    class Meta:
        unique_together = ['frist', 'projekt_zuweisung', 'arbeits_tag']

class Abgeschlossen(models.Model):
    projekt_zuweisung = models.ForeignKey(
        'ProjektZuweisung',
        on_delete=models.RESTRICT
    )
    frist = models.ForeignKey(
        'Frist',
        on_delete=models.RESTRICT,
        related_name='abgeschlossen'
    )
    ist_abgeschlossen = models.BooleanField(
        default=False
    )

    class Meta:
        unique_together= ['projekt_zuweisung', 'frist']
