import datetime
from django.shortcuts import get_object_or_404, render
from django.db.models import Q
from django.views.generic import TemplateView
from django.utils.timezone import make_aware
from django.db.models import Sum
from .. import forms
from .. import models
from .alle import FristAuswahlView

class MitarbeiterInFristAuswahlView(FristAuswahlView):
    pattern_name = 'arbeitszeit:mitarbeiter_in_stundenkonto'

class MitarbeiterInStundenkontoView(TemplateView):
    template_name = 'arbeitszeit/mitarbeiter_in_stundenkonto.html'
    frist = None
    projekt = None
    mitarbeiter_zuweisung = None
    stundenkonto = None
    abgeschlossen = None
    jetzt = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.frist = get_object_or_404(
            models.Frist,
            pk=kwargs.get('frist')
        )
        self.projekt = get_object_or_404(
            models.Projekt,
            kostenstelle=kwargs.get('kostenstelle')
        )
        self.mitarbeiter_zuweisung = get_object_or_404(
            models.ProjektZuweisung,
            projekt=self.projekt,
            nutzer_in=self.request.user
        )
        self.stundenkonto = models.Stundenkonto.objects.filter(
            projekt_zuweisung=self.mitarbeiter_zuweisung,
            frist=self.frist
        )
        self.jetzt = make_aware(datetime.datetime.now())
        if self.frist.frist_ende >= self.jetzt:
            self.abgeschlossen = models.Abgeschlossen.objects.get_or_create(
                frist=self.frist,
                projekt_zuweisung=self.mitarbeiter_zuweisung,
                defaults={'ist_abgeschlossen' : False}
            )[0]

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['frist'] = self.frist
        context['projekt'] = self.projekt
        context['abgeschlossen'] = self.abgeschlossen
        # Fristenauswahl
        fristen = models.Frist.objects.filter(
            Q(abgeschlossen__projekt_zuweisung=self.mitarbeiter_zuweisung)
            | Q(frist_ende__gte=self.jetzt)
        ).distinct()
        context['frist_auswahl'] = forms.FristAuswahlForm(
            fristen=fristen,
            anfang=self.frist.pk
        )
        context['abschliessen_form'] = forms.AbschliessenForm()
        return context

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(*args, **kwargs)
        if self.abgeschlossen is None or self.abgeschlossen.ist_abgeschlossen:
            context['ist_abgeschlossen'] = True
            context['stundenkonto'] = self.stundenkonto
            summe = self.stundenkonto.aggregate(Sum('arbeits_zeit')).get(
                'arbeits_zeit__sum'
            )
            context['summe'] = summe or 0
        else:
            context['ist_abgeschlossen'] = self.abgeschlossen.ist_abgeschlossen
            initial = {
                str(eintrag.arbeits_tag) :
                eintrag.arbeits_zeit for eintrag in self.stundenkonto
            }
            context['kalender'] = forms.KalenderForm(
                frist=self.frist,
                ist_abgeschlossen=self.abgeschlossen.ist_abgeschlossen,
                zuweisung=self.mitarbeiter_zuweisung,
                initial=initial
            )
        return render(
                request=request,
                template_name=self.template_name,
                context=context
            )

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(*args, **kwargs)
        if request.POST.get('abschliessen') is not None:
            self.abgeschlossen.ist_abgeschlossen = True
            self.abgeschlossen.save()
            context['stundenkonto'] = self.stundenkonto
            context['ist_zusammenfassung'] = True
            context['ist_abgeschlossen'] = True
            summe = self.stundenkonto.aggregate(Sum('arbeits_zeit')).get(
                    'arbeits_zeit__sum'
                )
            context['summe'] = summe or 0
        else:
            initial = {
                    str(eintrag.arbeits_tag) :
                    eintrag.arbeits_zeit for eintrag in self.stundenkonto
                }
            kalender = forms.KalenderForm(
                self.frist,
                self.mitarbeiter_zuweisung,
                self.abgeschlossen.ist_abgeschlossen,
                self.request.POST,
                initial=initial,
            )
            if kalender.is_valid():
                kalender.save()
                context['stundenkonto'] = models.Stundenkonto.objects.filter(
                    projekt_zuweisung=self.mitarbeiter_zuweisung,
                    frist=self.frist
                )
                summe = self.stundenkonto.aggregate(Sum('arbeits_zeit')).get(
                    'arbeits_zeit__sum'
                )
                context['summe'] = summe or 0
                context['ist_zusammenfassung'] = True
            else:
                context['kalender'] = kalender
        return render(
            request=request,
            template_name=self.template_name,
            context=context
        )
