from functools import reduce
from smtplib import SMTPException
from django.urls import reverse_lazy
from django.http import Http404
from django.contrib.auth import get_user_model
from django.core.mail import send_mail
from django.db.models import Q
from django.views.generic import TemplateView, FormView, UpdateView
from django.shortcuts import get_object_or_404, render, redirect
from django.core.exceptions import PermissionDenied
from django.template.loader import get_template
from nutzerverwaltung.models import NeueRNutzerIn
from .. import forms
from .. import models
from .alle import FristAuswahlView, SuchView

class ProjektleiterInUebersichtView(TemplateView):
    template_name = 'arbeitszeit/projekt/projektleiter_in_uebersicht.html'
    projekt = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        kostenstelle = kwargs.get('kostenstelle')
        self.projekt = get_object_or_404(
            models.Projekt,
            kostenstelle=kostenstelle
        )

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['projekt'] = self.projekt

        projektleiter_in_zuweisungen = models.ProjektLeiterIn.objects.filter(
            leitet=self.projekt
        )
        projektleiter_in_zuweisungen.select_related('nutzer_in')
        projektleiter_in_zuweisungen = projektleiter_in_zuweisungen.order_by(
            '-nutzer_in__is_active',
            'nutzer_in__nachname',
            'nutzer_in__vorname'
        )
        context['projektleiter_in_zuweisungen'] = projektleiter_in_zuweisungen

        mitarbeiter_in_zuweisungen = models.ProjektZuweisung.objects.filter(
            projekt=self.projekt
        )
        mitarbeiter_in_zuweisungen.select_related('nutzer_in')
        mitarbeiter_in_zuweisungen = mitarbeiter_in_zuweisungen.order_by(
            '-nutzer_in__is_active',
            'nutzer_in__nachname',
            'nutzer_in__vorname'
        )
        context['mitarbeiter_in_zuweisungen'] = mitarbeiter_in_zuweisungen

        return context

class ProjektMitarbeiterInSuchen(TemplateView):
    template_name = 'arbeitszeit/projekt/projekt_mitarbeiter_in_suchen.html'
    projekt = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.projekt = get_object_or_404(
            models.Projekt,
            kostenstelle=kwargs.get('kostenstelle')
        )

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['projekt'] = self.projekt
        context['suchformular'] = forms.MitarbeiterInSuchForm()
        return context

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(*args, **kwargs)
        suchformular = forms.MitarbeiterInSuchForm(request.POST)
        if suchformular.is_valid():
            reduce_func = lambda a, b: a | b
            map_func = lambda s: Q(vorname__icontains=s) | Q(
                nachname__icontains=s)
            eingabe = suchformular.cleaned_data.get('suchfeld').split(' ')
            suche = get_user_model().objects.exclude(
                projektzuweisung__projekt=self.projekt
            )
            suche = suche.filter(
                reduce(reduce_func, map(map_func, eingabe))
            )
            suche = suche.order_by('nachname', 'vorname')
            context['gefundene_mitarbeiter_innen'] = suche
        else:
            context['suchfeld'] = suchformular
        return render(
            request=request,
            template_name=self.template_name,
            context=context
        )

class ProjektMitarbeiterInHinzufuegenView(FormView):
    template_name = 'arbeitszeit/projekt/projekt_mitarbeiter_hinzufuegen.html'
    form_class = forms.MitarbeiterInGehaltsstufenForm
    email_template = 'arbeitszeit/projekt/projekt_mitarbeiter_in_hinzufuegen_mail.txt'
    success_url = None
    projekt = None
    nutzer_in = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.projekt = get_object_or_404(
            models.Projekt,
            kostenstelle=kwargs.get('kostenstelle')
        )
        self.nutzer_in = get_object_or_404(
            get_user_model(),
            pk=kwargs['nutzer_in']
        )
        self.success_url = reverse_lazy(
            'arbeitszeit:projekt_mitarbeiter_in_hinzufuegen_erfolg',
            kwargs={'kostenstelle': kwargs.get('kostenstelle')}
        )

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['nutzer_in'] = self.nutzer_in
        context['projekt'] = self.projekt
        return context

    def form_valid(self, form):
        gehalts_stufe = get_object_or_404(
            models.Gehaltsstufe,
            pk=form.cleaned_data.get('stufen_auswahl')
        )
        zuweisung = models.ProjektZuweisung.objects.create(
            nutzer_in=self.nutzer_in,
            hat_gehalts_stufe=gehalts_stufe,
            projekt=self.projekt
        )
        template = get_template(self.email_template)
        email_context = {
            'vorname': self.nutzer_in.vorname,
            'nachname': self.nutzer_in.nachname,
            'projekt': self.projekt.projekt_name
        }
        try:
            send_mail(
                subject='Zuordnung',
                message=template.render(context=email_context),
                from_email=None,
                recipient_list=[self.nutzer_in.email]
            )
        except SMTPException as exc:
            zuweisung.delete()
            raise SMTPException from exc

        return super().form_valid(form)

class ProjektMitarbeiterInHinzufuegenErfolgView(TemplateView):
    template_name = 'arbeitszeit/projekt/projekt_mitarbeiter_hinzufuegen_erfolg.html'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['projekt'] = get_object_or_404(
            models.Projekt,
            kostenstelle=kwargs.get('kostenstelle')
        )
        return context

class ProjektMitarbeiterInRegistrierenView(FormView):
    template_name = 'arbeitszeit/projekt/projekt_mitarbeiter_in_registrieren.html'
    form_class = forms.ProjektMitarbeiterRegistrierungForm
    email_template = 'arbeitszeit/projekt/registrierungs_mail.txt'

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.success_url = reverse_lazy(
            'arbeitszeit:projekt_mitarbeiter_in_registrieren_erfolg',
            kwargs={'kostenstelle': kwargs.get('kostenstelle')}
        )

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['kostenstelle'] = self.kwargs.get('kostenstelle')
        return context

    def form_valid(self, form):
        neue_nutzer_in = form.save()
        vermerk = NeueRNutzerIn.objects.create(nutzer_in=neue_nutzer_in)
        signed_token = vermerk.generiere_signierten_link(erneuern=False)

        template = get_template(self.email_template)
        email_context = {
            'signed_token': signed_token,
            'vorname': neue_nutzer_in.vorname,
            'nachname': neue_nutzer_in.nachname
        }
        try:
            send_mail(
                subject='Registrierung',
                message=template.render(context=email_context),
                from_email=None,
                recipient_list=[neue_nutzer_in.email]
            )
        except SMTPException as exc:
            vermerk.delete()
            neue_nutzer_in.delete()
            raise SMTPException from exc

        projekt = get_object_or_404(
            models.Projekt,
            kostenstelle=self.kwargs.get('kostenstelle')
        )
        gehaltsstufe = get_object_or_404(
            models.Gehaltsstufe,
            pk=form.cleaned_data.get('stufen_auswahl')
        )
        models.ProjektZuweisung.objects.create(
            nutzer_in=neue_nutzer_in,
            hat_gehalts_stufe=gehaltsstufe,
            projekt=projekt
        )
        return super().form_valid(form)

class ProjektMitarbeiterInRegistrierenErfolgView(TemplateView):
    template_name = 'arbeitszeit/projekt/projekt_mitarbeiter_in_registrieren_erfolg.html'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['kostenstelle'] = kwargs.get('kostenstelle')
        return context

class ProjektDatenAendernView(UpdateView):
    form_class = forms.ProjektAendernForm
    template_name = 'arbeitszeit/projekt/projekt_daten_aendern.html'

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.success_url = reverse_lazy(
            'arbeitszeit:projektleiter_in_uebersicht',
            kwargs={'kostenstelle':kwargs.get('kostenstelle')}
        )

    def get_queryset(self):
        return self.form_class.Meta.model.objects

    def get_object(self, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()
        queryset = queryset.filter(
            kostenstelle=self.kwargs.get('kostenstelle')
        )
        try:
            obj = queryset.get()
        except queryset.model.DoesNotExist:
            raise Http404 from obj
        return obj

    def get_form(self, form_class=None):
        form = super().get_form(form_class)
        form.fields['laufzeit_anfang'].widget = forms.DateInput()
        form.fields['laufzeit_ende'].widget = forms.DateInput()
        return form

class ProjektEinrichtungAuswahlView(SuchView):
    model = models.Einrichtung
    such_felder = ['einrichtungs_name']
    ordering = ['einrichtungs_name']
    template_name = 'arbeitszeit/verwaltung/projekt_einrichtung_auswaehlen.html'

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.success_url = reverse_lazy(
            'arbeitszeit:projektleiter_in_uebersicht',
            kwargs={'kostenstelle':kwargs.get('kostenstelle')}
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["einrichtungs_auswahl"] = forms.ProjektEinrichtungsAuswahlForm(
            self.get_queryset()
        )
        return context
    
    def post(self, request, *args, **kwargs):
        form = forms.ProjektEinrichtungsAuswahlForm(
            self.get_queryset(),
            self.request.POST
        )
        if form.is_valid():
            einrichtung_pk = form.cleaned_data['einrichtungen']
            einrichtung = models.Einrichtung.objects.filter(pk=einrichtung_pk)
            if einrichtung.exists():
                models.Projekt.objects.filter(
                    kostenstelle=kwargs['kostenstelle']
                ).update(
                    einrichtung=einrichtung[0]
                )
                return redirect(
                    self.success_url,
                    kostenstelle=self.kwargs['kostenstelle']
                )
        
        return self.get(request, *args, **kwargs)

class ProjektMitarbeiterInAendernView(TemplateView):
    template_name = 'arbeitszeit/projekt/projekt_mitarbeiter_in_aendern.html'
    zuweisung = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        try:
            self.zuweisung = models.ProjektZuweisung.objects.select_related(
                'hat_gehalts_stufe',
                'nutzer_in',
            ).get(
                projekt__kostenstelle=kwargs.get('kostenstelle'),
                pk=kwargs.get('pk'),
            )
        except models.ProjektZuweisung.DoesNotExist:
            raise PermissionDenied from self.zuweisung

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['nutzer_in_daten'] = forms.ProjektMitarbeiterForm(
            instance=self.zuweisung.nutzer_in
        )
        context['zuweisung'] = forms.ProjektZuweisungForm(
            instance=self.zuweisung
        )
        return context

    def post(self, request, *args, **kwargs):
        nutzer_in_daten = forms.ProjektMitarbeiterForm(
            request.POST,
            instance=self.zuweisung.nutzer_in
        )
        zuweisung = forms.ProjektZuweisungForm(
            request.POST,
            instance=self.zuweisung
        )
        if nutzer_in_daten.is_valid() and zuweisung.is_valid():
            nutzer_in_daten.save()
            zuweisung.save()
            return redirect(
                'arbeitszeit:projektleiter_in_uebersicht',
                kwargs.get('kostenstelle')
            )
        else:
            context = {
                'nutzer_in_daten': nutzer_in_daten,
                'zuweisung': zuweisung,
            }
            return render(
                request=request,
                template_name=self.template_name,
                context=context
            )

class ProjektFristAuswahlView(FristAuswahlView):
    pattern_name = 'arbeitszeit:projekt_stundenzettel'

class ProjektStundenzettelView(TemplateView):
    template_name = 'arbeitszeit/projekt/stundenzettel.html'
    projekt = None
    frist = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.projekt = get_object_or_404(
            models.Projekt,
            kostenstelle=kwargs.get('kostenstelle')
        )
        self.frist = get_object_or_404(
            models.Frist,
            pk=kwargs.get('frist')
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['frist_auswahl'] = forms.FristAuswahlForm(
            models.Frist.objects.all(),
            self.frist.pk
        )
        return context
    

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        context['form'] = forms.Stundenzettel(
            self.projekt,
            self.frist
        )
        return render(
            request,
            template_name=self.template_name,
            context=context
        )

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        form = forms.Stundenzettel(
            self.projekt,
            self.frist,
            data=request.POST
        )
        context['form'] = form
        if form.is_valid():
            form.save()
        return render(
            request,
            template_name=self.template_name,
            context=context
        )
    