from django.views.generic import (TemplateView, FormView,
    ListView, CreateView, UpdateView, DeleteView, DetailView)
from django.urls import reverse_lazy, reverse
from django.shortcuts import redirect
from django.shortcuts import get_object_or_404, render
from django.template.loader import get_template
from django.core.mail import send_mail
from smtplib import SMTPException
from .. import forms
from .alle import SuchView, FristAuswahlView
from .projektleiterin import (
    ProjektleiterInUebersichtView,
    ProjektDatenAendernView,
    ProjektEinrichtungAuswahlView,
    ProjektStundenzettelView
)
from nutzerverwaltung.models import NutzerIn, NeueRNutzerIn
from arbeitszeit.models import Projekt, Einrichtung, Gehaltsstufe, ProjektLeiterIn, ProjektZuweisung, Frist
from eigentags.models import SystemNachricht
from urllib.parse import quote_plus

class VerwalterInUebersichtView(TemplateView):
    template_name = 'arbeitszeit/verwaltung/uebersicht.html'

class VerwalterNutzerInSucheView(SuchView):
    model = NutzerIn
    such_felder = ['nachname', 'vorname', 'email']
    ordering = ['nachname', 'vorname']
    template_name = 'arbeitszeit/verwaltung/nutzer_in_suche.html'

class VerwalterNutzerInDetailsView(DetailView):
    template_name = 'arbeitszeit/verwaltung/nutzer_in_details.html'
    model = NutzerIn

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['mitarbeiter_von'] = ProjektZuweisung.objects.filter(
            nutzer_in=self.object
        ).select_related('projekt').select_related('hat_gehalts_stufe')
        context['leitet_projekte'] = ProjektLeiterIn.objects.filter(
            nutzer_in=self.object).select_related('leitet')
        return context 

class VerwalterNutzerInAendernView(UpdateView):
    model = NutzerIn
    fields = [
        'email',
        'nachname',
        'vorname',
        'is_active',
        'is_verwalter',
    ]
    template_name = 'arbeitszeit/verwaltung/nutzer_in_aendern.html'

    def get_success_url(self, **kwargs):
        return reverse(
            'arbeitszeit:nutzer_in_details', 
            kwargs = {'pk' : self.kwargs['pk']}
        ) 

class VerwalterNutzerInRegistrierenView(FormView):
    form_class= forms.MitarbeiterRegistrierungForm
    template_name = 'arbeitszeit/verwaltung/nutzer_in_registrieren.html'
    email_template = 'arbeitszeit/verwaltung/nutzer_in_registrieren_email.txt'
    success_url = reverse_lazy('arbeitszeit:neue_nutzer_innen')

    def form_valid(self, form):
        neue_nutzer_in = form.save()
        vermerk = NeueRNutzerIn.objects.create(nutzer_in=neue_nutzer_in)
        signed_token = vermerk.generiere_signierten_link(erneuern=False)

        template = get_template(self.email_template)
        email_context = {
            'signed_token': signed_token,
            'vorname': neue_nutzer_in.vorname,
            'nachname': neue_nutzer_in.nachname
        }
        try:
            send_mail(
                subject='Registrierung',
                message=template.render(context=email_context),
                from_email=None,
                recipient_list=[neue_nutzer_in.email]
            )
        except SMTPException as exc:
            vermerk.delete()
            neue_nutzer_in.delete()
            raise SMTPException from exc

        return super().form_valid(form)

class VerwalterNeueRNutzerInView(ListView):
    model = NeueRNutzerIn
    paginate_by = 25
    ordering = ['nutzer_in__nachname', 'nutzer_in__vorname']
    template_name = 'arbeitszeit/verwaltung/neue_nutzer_in.html'

class VerwalterNeueRNutzerInErneuernView(UpdateView):
    model = NeueRNutzerIn
    fields = []
    template_name = 'arbeitszeit/verwaltung/neue_nutzer_in_erneuern.html'
    success_url = reverse_lazy('arbeitszeit:neue_nutzer_innen')
    email_template = 'arbeitszeit/verwaltung/nutzer_in_registrieren_email_erneuert.txt'

    def form_valid(self, form):
        signed_token = self.object.generiere_signierten_link(erneuern=True)

        template = get_template(self.email_template)
        email_context = {
            'signed_token': signed_token,
            'vorname': self.object.nutzer_in.vorname,
            'nachname': self.object.nutzer_in.nachname
        }
        try:
            send_mail(
                subject='Registrierung',
                message=template.render(context=email_context),
                from_email=None,
                recipient_list=[self.object.nutzer_in.email]
            )
        except SMTPException as exc:
            vermerk.delete()
            neue_nutzer_in.delete()
            raise SMTPException from exc

        return super().form_valid(form)

class VerwalterProjektSucheView(SuchView):
    model = Projekt
    such_felder = ['projekt_name', 'kostenstelle', 'auftraggeber']
    ordering = ['kostenstelle']
    template_name = 'arbeitszeit/verwaltung/projekt_suche.html'

class VerwalterProjektErstellenView(CreateView):
    form_class = forms.ProjektErstellenForm
    template_name = 'arbeitszeit/verwaltung/projekt_erstellen.html'

    def get_form(self, form_class=None):
        form = super().get_form(form_class)
        form.fields['laufzeit_anfang'].widget = forms.DateInput()
        form.fields['laufzeit_ende'].widget = forms.DateInput()
        return form
    
    def form_valid(self, form):
        kostenstelle = form.cleaned_data['kostenstelle']
        self.success_url = reverse(
            'arbeitszeit:projekt_einrichtung_auswaehlen',
            kwargs={'kostenstelle':kostenstelle}
        )
        return super().form_valid(form)

class VerwalterProjektUebersichtView(ProjektleiterInUebersichtView):
    template_name = 'arbeitszeit/verwaltung/projekt_uebersicht.html'

class VerwalterProjektEinrichtungAuswahlView(ProjektEinrichtungAuswahlView):
    template_name = 'arbeitszeit/verwaltung/projekt_einrichtung_auswaehlen.html'

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.success_url = reverse_lazy(
            'arbeitszeit:projekt_uebersicht',
            kwargs={'kostenstelle':kwargs.get('kostenstelle')}
        )

class VerwalterProjektAendernView(ProjektDatenAendernView):
    template_name = 'arbeitszeit/verwaltung/projekt_daten_aendern.html'

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.success_url = reverse_lazy(
            'arbeitszeit:projekt_uebersicht',
            kwargs={'kostenstelle':kwargs.get('kostenstelle')}
        )
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['kostenstelle'] = self.kwargs['kostenstelle']
        return context

class VerwalterProjektleiterInSuchen(SuchView):
    model = NutzerIn
    such_felder = ['nachname', 'vorname', 'email']
    ordering = ['nachname', 'vorname']
    template_name = 'arbeitszeit/verwaltung/projekt_leiter_in_suchen.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['kostenstelle'] = self.kwargs['kostenstelle']
        return context

class VerwalterProjektleiterInHinzufuegen(CreateView):
    model = ProjektLeiterIn
    form_class = forms.ProjektLeiterInForm
    template_name = 'arbeitszeit/verwaltung/projekt_leiter_in_hinzufuegen.html'

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.projekt = get_object_or_404(
            Projekt,
            kostenstelle=self.kwargs['kostenstelle']
        )
        self.nutzer_in = get_object_or_404(
            NutzerIn,
            pk=self.kwargs['nutzer_in_pk']
        )

    def get_initial(self):
        initial = super().get_initial()
        initial['leitet'] = self.projekt.pk
        initial['nutzer_in'] = self.nutzer_in.pk
        initial['von'] = self.projekt.laufzeit_anfang
        initial['bis'] = self.projekt.laufzeit_ende
        return initial

    def get_success_url(self, **kwargs):
        return reverse(
            'arbeitszeit:projekt_uebersicht', 
            kwargs = {'kostenstelle' : self.kwargs['kostenstelle']}
        )
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['kostenstelle'] = self.kwargs['kostenstelle']
        context['nutzer_in'] = self.nutzer_in
        context['projekt'] = self.projekt
        return context
    
class VerwalterProjektleiterInAendern(UpdateView):
    model = ProjektLeiterIn
    form_class = forms.ProjektLeiterInForm
    template_name = 'arbeitszeit/verwaltung/projekt_leiter_in_aendern.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['kostenstelle'] = self.kwargs['kostenstelle']
        return context

    def get_success_url(self, **kwargs):
        return reverse(
            'arbeitszeit:projekt_uebersicht', 
            kwargs = {'kostenstelle' : self.kwargs['kostenstelle']}
        )

class VerwalterProjektStundenzettelView(ProjektStundenzettelView):
    template_name = 'arbeitszeit/verwaltung/stundenzettel.html'

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        context['form'] = forms.Stundenzettel(
            self.projekt,
            self.frist,
            True
        )
        return render(
            request,
            template_name=self.template_name,
            context=context
        )

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        form = forms.Stundenzettel(
            self.projekt,
            self.frist,
            True,
            request.POST
        )
        context['form'] = form
        if form.is_valid():
            form.save()
        return render(
            request,
            template_name=self.template_name,
            context=context
        )

class VerwalterProjektFristAuswahlView(FristAuswahlView):
    pattern_name = 'arbeitszeit:verwalter_projekt_stundenzettel'

class VerwalterMitarbeiterInSuchen(SuchView):
    model = NutzerIn
    such_felder = ['nachname', 'vorname', 'email']
    ordering = ['nachname', 'vorname']
    template_name = 'arbeitszeit/verwaltung/verwalter_projekt_mitarbeiter_in_suchen.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['kostenstelle'] = self.kwargs['kostenstelle']
        return context

class VerwalterMitarbeiterInHinzufuegen(CreateView):
    model = ProjektZuweisung
    form_class = forms.VerwaltungProjektMitarbeiterInForm
    template_name = 'arbeitszeit/verwaltung/projekt_mitarbeiter_in_hinzufuegen.html'

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.projekt = get_object_or_404(
            Projekt,
            kostenstelle=self.kwargs['kostenstelle']
        )
        self.nutzer_in = get_object_or_404(
            NutzerIn,
            pk=self.kwargs['nutzer_in_pk']
        )

    def get_initial(self):
        initial = super().get_initial()
        initial['leitet'] = self.projekt.pk
        initial['nutzer_in'] = self.nutzer_in.pk
        return initial

    def get_success_url(self, **kwargs):
        return reverse(
            'arbeitszeit:projekt_uebersicht', 
            kwargs = {'kostenstelle' : self.kwargs['kostenstelle']}
        )
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['kostenstelle'] = self.kwargs['kostenstelle']
        context['nutzer_in'] = self.nutzer_in
        context['projekt'] = self.projekt
        return context

class VerwalterMitarbeiterInAendern(UpdateView):
    model = ProjektZuweisung
    form_class = forms.VerwaltungProjektMitarbeiterInForm
    template_name = 'arbeitszeit/verwaltung/projekt_mitarbeiter_in_aendern.html'

    def get_success_url(self, **kwargs):
        return reverse(
            'arbeitszeit:projekt_uebersicht', 
            kwargs = {'kostenstelle' : self.kwargs['kostenstelle']}
        )

class VerwalterEinrichtungsSucheView(SuchView):
    model = Einrichtung
    such_felder = ['einrichtungs_name']
    ordering = ['einrichtungs_name']
    template_name = 'arbeitszeit/verwaltung/einrichtungs_suche.html'

class VerwalterEinrichtungErstellenView(CreateView):
    model = Einrichtung
    fields = ['einrichtungs_name', 'verbuchungs_stelle']
    template_name = 'arbeitszeit/verwaltung/einrichtung_erstellen.html'

    def get_success_url(self):
        url = reverse('arbeitszeit:einrichtungs_suche')
        suche = quote_plus(self.request.POST['einrichtungs_name'])
        return f"{url}?suchfeld={suche}"

class VerwalterEinrichtungAendernView(UpdateView):
    model = Einrichtung
    fields = ['einrichtungs_name', 'verbuchungs_stelle']
    template_name = 'arbeitszeit/verwaltung/einrichtung_aendern.html'

    def get_success_url(self):
        url = reverse('arbeitszeit:einrichtungs_suche')
        suche = quote_plus(self.request.POST['einrichtungs_name'])
        return f"{url}?suchfeld={suche}"

class VerwalterEinrichtungLoeschenView(DeleteView):
    model = Einrichtung
    template_name = 'arbeitszeit/verwaltung/einrichtung_loeschen.html'

    def get_success_url(self):
        url = reverse('arbeitszeit:einrichtungs_suche')
        suche = quote_plus(self.model.objects.filter(pk=self.kwargs['pk']).first().einrichtungs_name)
        return f"{url}?suchfeld={suche}"

class VerwalterGehaltsstufenView(ListView):
    model = Gehaltsstufe
    template_name = 'arbeitszeit/verwaltung/gehaltsstufen.html'

class VerwalterGehaltsstufeErstellenView(CreateView):
    model = Gehaltsstufe
    fields = ['stufen_kuerzel']
    template_name = 'arbeitszeit/verwaltung/gehaltsstufe_erstellen.html'
    success_url = reverse_lazy('arbeitszeit:gehaltsstufen')

class VerwalterGehaltsstufeAendernView(UpdateView):
    model = Gehaltsstufe
    fields = ['stufen_kuerzel']
    template_name = 'arbeitszeit/verwaltung/gehaltsstufe_aendern.html'
    success_url = reverse_lazy('arbeitszeit:gehaltsstufen')

class VerwalterGehaltsstufeLoeschenView(DeleteView):
    model = Gehaltsstufe
    template_name = 'arbeitszeit/verwaltung/gehaltsstufe_loeschen.html'
    success_url = reverse_lazy('arbeitszeit:gehaltsstufen')

class VerwalterFristView(ListView):
    model = Frist
    template_name = 'arbeitszeit/verwaltung/fristen.html'

class VerwalterFristAendernView(UpdateView):
    model = Frist
    fields = '__all__'
    template_name = 'arbeitszeit/verwaltung/fristen_aendern.html'
    success_url = reverse_lazy('arbeitszeit:fristen')

    def get_form(self, form_class=None):
        form = super().get_form(form_class)
        form.fields['zeitraum_anfang'].widget = forms.DateInput()
        form.fields['zeitraum_ende'].widget = forms.DateInput()
        form.fields['frist_ende'].widget = forms.DateTimeInput()
        return form

class VerwalterSystemNachrichtenView(ListView):
    model = SystemNachricht
    template_name = 'arbeitszeit/verwaltung/systemnachrichten.html'

class VerwalterSystemNachrichtenAendernView(UpdateView):
    model = SystemNachricht
    fields = '__all__'
    template_name = 'arbeitszeit/verwaltung/systemnachricht_aendern.html'
    success_url = reverse_lazy('arbeitszeit:systemnachrichten')

class VerwalterSystemNachrichtenLoeschenView(DeleteView):
    model = SystemNachricht
    template_name = 'arbeitszeit/verwaltung/systemnachricht_loeschen.html'
    success_url = reverse_lazy('arbeitszeit:systemnachrichten')

class VerwalterSystemNachrichtenErstellenView(CreateView):
    model = SystemNachricht
    fields = '__all__'
    template_name = 'arbeitszeit/verwaltung/systemnachricht_erstellen.html'
    success_url = reverse_lazy('arbeitszeit:systemnachrichten')
